#!/usr/bin/env python3

import struct, hexdump, hashlib, binascii

# Stolen from https://codereview.stackexchange.com/questions/37648/python-implementation-of-sha1
# And reworked
def sha1_core(init_hash, padded_data):
    print ("sha1_core(), init_hash", init_hash)
    print ("sha1_core(), loading state", init_hash)
    assert len(init_hash)==8*5
    idx=0
    h0=int(init_hash[idx:idx+8], 16)
    idx=idx+8
    h1=int(init_hash[idx:idx+8], 16)
    idx=idx+8
    h2=int(init_hash[idx:idx+8], 16)
    idx=idx+8
    h3=int(init_hash[idx:idx+8], 16)
    idx=idx+8
    h4=int(init_hash[idx:idx+8], 16)
    assert '%08x%08x%08x%08x%08x' % (h0, h1, h2, h3, h4) == init_hash

    def rol(n, b):
        return ((n << b) | (n >> (32 - b))) & 0xffffffff

    print ("sha1_core(), padded_data:")
    hexdump.hexdump(padded_data)
    print ("sha1_core(), len(padded_data)", len(padded_data))

    thunks = [padded_data[i:i+64] for i in range(0, len(padded_data), 64)]
    for thunk in thunks:
        w = list(struct.unpack('>16L', thunk)) + [0] * 64
        for i in range(16, 80):
            w[i] = rol((w[i-3] ^ w[i-8] ^ w[i-14] ^ w[i-16]), 1)

        a, b, c, d, e = h0, h1, h2, h3, h4

        # Main loop
        for i in range(0, 80):
            if 0 <= i < 20:
                f = (b & c) | ((~b) & d)
                k = 0x5A827999
            elif 20 <= i < 40:
                f = b ^ c ^ d
                k = 0x6ED9EBA1
            elif 40 <= i < 60:
                f = (b & c) | (b & d) | (c & d) 
                k = 0x8F1BBCDC
            elif 60 <= i < 80:
                f = b ^ c ^ d
                k = 0xCA62C1D6

            a, b, c, d, e = rol(a, 5) + f + e + k + w[i] & 0xffffffff, \
                            a, rol(b, 30), c, d

        h0 = h0 + a & 0xffffffff
        h1 = h1 + b & 0xffffffff
        h2 = h2 + c & 0xffffffff
        h3 = h3 + d & 0xffffffff
        h4 = h4 + e & 0xffffffff

    result='%08x%08x%08x%08x%08x' % (h0, h1, h2, h3, h4)
    print ("sha1_core() result", result)
    return result

# from http://docs.oracle.com/javase/specs/jvms/se7/html/jvms-3.html (3.3)
def align2grain (i, grain):
    return ((i + grain-1) & ~(grain-1))

def calc_pad_bytes_to_add(size):
    return align2grain(size, 64)-size

def add_padding(data, size):
    return data + b"\x80" + b"\x00"*(calc_pad_bytes_to_add(size)-(8+1)) + struct.pack('>Q', 8 * size)

def sha1(init_hash, data):
    return sha1_core(init_hash, add_padding(data, len(data)))

sha1_standard_init_state="67452301efcdab8998badcfe10325476c3d2e1f0"

# self-test
#"""
# TODO try longer strings
TEST_MSG=b"hello, world"
m=hashlib.sha1()
buf_to_hash=TEST_MSG
hexdump.hexdump(buf_to_hash)
m.update(buf_to_hash)
hash1=m.hexdigest()

hash2=sha1(sha1_standard_init_state, TEST_MSG)
assert hash1==hash2
#"""

f=open("packet.bin", "rb")
buf=f.read()
f.close()

MAC_size=20 # SHA1

# Cut off MAC:
MAC_from_file=buf[-MAC_size:]
message=buf[0:len(buf)-MAC_size]

secret_key_len=12
# We know message length
# We also know that the secret key has 12 bytes
# So we can easily reconstruct the buffer that was passed to sha1_core()
# But not secret key. Here it is shown as a string of "???..."
original_buf_len=secret_key_len+len(message)
original_buf_padded=add_padding(b"?"*secret_key_len + message, original_buf_len)
print ("original_buf_padded")
hexdump.hexdump(original_buf_padded)

# We know that after hashing buf via sha1_core(), its state was MAC_from_file

MAC_from_file_hex=binascii.hexlify(MAC_from_file).decode("utf-8")

data_we_add=b"This is added by attacker."

# Load that state and hash the data we want to add
# But the final padding block must contain length of the whole buffer

original_buf_padded_plus_our_data_len=len(original_buf_padded)+len(data_we_add)

# As of padding: 8 bytes for final 64-bit word, 1 byte for 0x80, other bytes are 0x00
# 2nd argument: length in bytes. But it's embedded in padding block as a number of bits.
def construct_padding(padding_len, length_at_the_end):
    return b"\x80" + b"\x00"*(padding_len-(8+1)) + struct.pack('>Q', length_at_the_end*8)

pad_len=calc_pad_bytes_to_add(original_buf_padded_plus_our_data_len)
new_MAC=sha1_core(MAC_from_file_hex, data_we_add + construct_padding(pad_len, original_buf_padded_plus_our_data_len))

print ("new_MAC", new_MAC)

# Now construct a new message:
pad_len=calc_pad_bytes_to_add(secret_key_len+len(message))
original_key_and_msg_len=secret_key_len+len(message)
new_buf=message + construct_padding(pad_len, original_key_and_msg_len) + data_we_add

print ("new_buf")
hexdump.hexdump(new_buf)

f=open("packet.bin", "wb")
f.write(new_buf + bytes.fromhex(new_MAC))
f.close()

